import { Component, OnInit, AfterContentInit } from '@angular/core';
import * as BABYLON from 'babylonjs';
import { SolidParticle } from 'babylonjs';


@Component({
  selector: 'app-test-bjs',
  templateUrl: './testBJS.component.html',
  styleUrls: ['./testBJS.component.css']
})
export class TestBJSComponent implements AfterContentInit {

  private _canvas: any; //HTMLCanvasElement;
  private _engine: BABYLON.Engine;
  private _scene: BABYLON.Scene;
  private _camera: BABYLON.ArcRotateCamera;
  private _light: BABYLON.Light;
  private _container: BABYLON.Mesh;

  private spaceSize = 100;

  //Base particle
  private particleRadius = 3;
  private sphere: BABYLON.Mesh;
  private particleMaxSpeed = 0.5; //distance per frame
  private particleMaxAngular = 0.025; //rotation per frame

  private leftNorm: BABYLON.Vector3;
  private leftPosition: BABYLON.Vector3;

  private rightNorm: BABYLON.Vector3;
  private rightPosition: BABYLON.Vector3;

  private frontNorm: BABYLON.Vector3;
  private frontPosition: BABYLON.Vector3;

  private backNorm: BABYLON.Vector3;
  private backPosition: BABYLON.Vector3;

  private groundNorm: BABYLON.Vector3;
  private groundPosition: BABYLON.Vector3;

  private roofNorm: BABYLON.Vector3;
  private roofPosition: BABYLON.Vector3;

  private offsets = [];

  constructor() {

  }

  ngAfterContentInit(): void {

    // Create the scene
    this.setupScene();

    // start animation
    this.animate();
  }

  setupScene(): void {

    // Create canvas and engine
    this._canvas = document.getElementById('renderCanvas');     // ok

    this.createEngine();     // ok

    // create a basic BJS Scene object
    this.createScene();     // ok

    // create a ArcRotateCamera
    this.createCamera();     // ok

    // create a basic light, aiming 0,1,0 - meaning, to the sky
    this.createLights();     // ok

    this.createContainer();     // ok

    this.createBasicSpheres();

    this.createSolidParticleSystem();


    //this.createParent();


  }

  createEngine() {     // ok
    this._engine = new BABYLON.Engine(this._canvas, true, { preserveDrawingBuffer: true, stencil: true });

    if (!this._engine) throw 'engine should not be null.';
  }

  rotateAboutAxisThroughPoint(vector3, axis, point, angle) {     // ok
    var _p = new BABYLON.Quaternion(vector3.x - point.x, vector3.y - point.y, vector3.z - point.z, 0);  //change _p for quaternion multiplication
    var q = BABYLON.Quaternion.RotationAxis(axis, angle);  //form quaternion rotation		
    var qinv = BABYLON.Quaternion.Inverse(q);
    var _pdash = q.multiply(_p).multiply(qinv);
    return new BABYLON.Vector3(_pdash.x + point.x, _pdash.y + point.y, _pdash.z + point.z); //change to Vector 3 for new position vector;
  }


  createScene() {     // ok
    this._scene = new BABYLON.Scene(this._engine);

    this._scene.collisionsEnabled = true;

    //Set gravity for the scene (G force like, on Y-axis)
    this._scene.gravity = new BABYLON.Vector3(0, -0.9, 0);
  }

  createCamera() {     // ok
    // create a ArcRotateCamera
    this._camera = new BABYLON.ArcRotateCamera("camera", -Math.PI / 2, Math.PI / 2, 200, BABYLON.Vector3.Zero(), this._scene);

    // target the camera to scene origin
    this._camera.setTarget(BABYLON.Vector3.Zero());

    // attach the camera to the canvas
    this._camera.attachControl(this._canvas, false);
  }

  createLights() {
    this._light = new BABYLON.HemisphericLight('light1', new BABYLON.Vector3(0, 1, 0), this._scene);
    this._light.intensity = 0.2;
    var pl = new BABYLON.PointLight('pl', this._camera.position, this._scene);
    pl.intensity = 0.3;
  }

  createContainer() {     // ok

    var ground = BABYLON.MeshBuilder.CreateGround("ground", { width: this.spaceSize, height: this.spaceSize }, this._scene);
    ground.position.y = -this.spaceSize / 2;

    var wallLeft = ground.clone("wallLeft");
    wallLeft.rotation.z = -Math.PI / 2;
    wallLeft.position.x = - this.spaceSize / 2;
    wallLeft.position.y = 0;

    var wallRight = ground.clone("wallRight");
    wallRight.rotation.z = Math.PI / 2;
    wallRight.position.x = this.spaceSize / 2;
    wallRight.position.y = 0;

    var wallFront = ground.clone("wallBack");
    wallFront.rotation.x = Math.PI / 2;
    wallFront.position.z = -this.spaceSize / 2;
    wallFront.position.y = 0;

    var wallBack = ground.clone("wallBack");
    wallBack.rotation.x = -Math.PI / 2;
    wallBack.position.z = this.spaceSize / 2;
    wallBack.position.y = 0;

    var roof = ground.clone("roof");
    roof.rotation.x = Math.PI;
    roof.position.y = this.spaceSize / 2;

    this.leftNorm = new BABYLON.Vector3(1, 0, 0);
    this.leftPosition = new BABYLON.Vector3(-this.spaceSize / 2, 0, 0);

    this.rightNorm = new BABYLON.Vector3(-1, 0, 0);
    this.rightPosition = new BABYLON.Vector3(this.spaceSize / 2, 0, 0);

    this.frontNorm = new BABYLON.Vector3(0, 0, 1);
    this.frontPosition = new BABYLON.Vector3(0, 0, -this.spaceSize / 2);

    this.backNorm = new BABYLON.Vector3(0, 0, -1);
    this.backPosition = new BABYLON.Vector3(0, 0, this.spaceSize / 2);

    this.groundNorm = new BABYLON.Vector3(0, 1, 0);
    this.groundPosition = new BABYLON.Vector3(0, -this.spaceSize / 2, 0);

    this.roofNorm = new BABYLON.Vector3(0, -1, 0);
    this.roofPosition = new BABYLON.Vector3(0, this.spaceSize / 2, 0);

    this.offsets[0] = new BABYLON.Vector3(1, 0, 0);
    this.offsets[1] = new BABYLON.Vector3(-1, 0, 0);
    this.offsets[2] = new BABYLON.Vector3(0, 1, 0);
    this.offsets[3] = new BABYLON.Vector3(0, -1, 0);
    this.offsets[4] = new BABYLON.Vector3(0, 0, 1);
    this.offsets[5] = new BABYLON.Vector3(0, 0, -1);

  }

  createBasicSpheres() {
    this.sphere = BABYLON.MeshBuilder.CreateSphere('sphere', { diameter: 2 * this.particleRadius, segments: 8 }, this._scene);
  }

  createSolidParticleSystem() {
    // Particle system
    var spheresPerShapeNb = 5;
    var particleNb = spheresPerShapeNb * 30;
    var SPS = new BABYLON.SolidParticleSystem('SPS', this._scene);
    SPS.addShape(this.sphere, particleNb);
    this.sphere.dispose();
    var mesh = SPS.buildMesh();
    var innerSpace = this.spaceSize - 2 * spheresPerShapeNb * this.particleRadius;

    // SPS initialization 
    SPS.initParticles = function () {
      for (var p = 0; p < SPS.nbParticles; p += spheresPerShapeNb) {
        var x = innerSpace * Math.random();
        var y = innerSpace * Math.random();
        var z = innerSpace * Math.random();
        let particle: SolidParticle = SPS.particles[p];

        particle.position = new BABYLON.Vector3(x - innerSpace / 2, y - innerSpace / 2, z - innerSpace / 2);
        particle.speed = this.particleMaxSpeed;//* Math.random();
        particle.direction = new BABYLON.Vector3(2 * Math.random() - 1, 2 * Math.random() - 1, 2 * Math.random() - 1).normalize();
        particle.velocity = particle.direction.scale(particle.speed);
        particle.color = new BABYLON.Color4(Math.random(), Math.random(), Math.random(), 1);
        particle.axis = new BABYLON.Vector3(Math.random() + 1, Math.random() + 1, Math.random() + 1);
        particle.angularSpeed = this.particleMaxAngular;//* Math.random();
        particle.centre = new BABYLON.Vector3(particle.position.x, particle.position.y, particle.position.z);
        particle.offset = BABYLON.Vector3.Zero();

        var moment = particle.centre;
        for (var i = 1; i < spheresPerShapeNb; i++) {
          var rnd = Math.floor(Math.random() * 6);
          let subParticle: any = SPS.particles[p + i];

          subParticle.offset = SPS.particles[p + i - 1].offset.add(this.offsets[rnd].scale(2 * this.particleRadius));
          subParticle.position = particle.position.add(subParticle.offset);
          moment.addInPlace(new BABYLON.Vector3(subParticle.position.x, subParticle.position.y, subParticle.position.z));
          subParticle.speed = particle.speed;
          subParticle.direction = particle.direction;
          subParticle.velocity = particle.velocity;
          subParticle.color = particle.color;
          subParticle.axis = particle.axis;
          subParticle.angularSpeed = particle.angularSpeed;
          subParticle.centre = particle.centre;
          subParticle.angle = 0;
        }
        moment.scaleToRef(1 / spheresPerShapeNb, particle.centre);
      }
    };

    var absNormX = 0;
    var absNormY = 0;
    var absNormZ = 0;
    var pCoord = 0;
    var sCoord = 0;
    var pCoordDir = 0;
    var sCoordDir = 0;
    var psDot = 0;

    var pNormalVec = BABYLON.Vector3.Zero();
    var pTangVec = BABYLON.Vector3.Zero();
    var normalVec = BABYLON.Vector3.Zero();
    var qNormalVec = BABYLON.Vector3.Zero();
    var qTangVec = BABYLON.Vector3.Zero();
    var pPos = BABYLON.Vector3.Zero();
    var pDir = BABYLON.Vector3.Zero();

    SPS.updateParticle = (particle: SolidParticle) => {
      var particleGroup = Math.floor(particle.idx / spheresPerShapeNb);
      var qGroup = 0;

      var setVelocity = function (sideNormal, sidePosition) {
        var impact = false;
        absNormX = Math.abs(sideNormal.x);
        absNormY = Math.abs(sideNormal.y);
        absNormZ = Math.abs(sideNormal.z);
        pCoord = particle.position.x * absNormX + particle.position.y * absNormY + particle.position.z * absNormZ;
        sCoord = sidePosition.x * absNormX + sidePosition.y * absNormY + sidePosition.z * absNormZ;
        pCoordDir = particle.direction.x * absNormX + particle.direction.y * absNormY + particle.direction.z * absNormZ;
        sCoordDir = sideNormal.x * absNormX + sideNormal.y * absNormY + sideNormal.z * absNormZ;
        if (pCoordDir * sCoordDir <= 0 && Math.abs(pCoord - sCoord) <= this.particleRadius) {
          sideNormal.scaleToRef(particle.direction.x * sideNormal.x + particle.direction.y * sideNormal.y + particle.direction.z * sideNormal.z, pNormalVec);
          particle.direction.subtractToRef(pNormalVec, pTangVec);
          pNormalVec.scaleInPlace(-1);
          pNormalVec.addToRef(pTangVec, particle.direction);
          particle.direction.normalize();
          particle.direction.scaleToRef(particle.speed, particle.velocity);
          for (var i = 0; i < spheresPerShapeNb; i++) {
            SPS.particles[spheresPerShapeNb * particleGroup + i].velocity = particle.velocity;
          }
          p = spheresPerShapeNb * particleGroup + spheresPerShapeNb - 1;
          impact = true;
        }
        return impact;
      }


      for (var p = (particleGroup + 1) * spheresPerShapeNb; p < SPS.nbParticles; p++) {
        var q = SPS.particles[p];
        qGroup = Math.floor(q.idx / spheresPerShapeNb);
        particle.position.subtractToRef(q.position, normalVec);
        particle.direction.subtractToRef(q.direction, pDir)
        if (BABYLON.Vector3.Dot(normalVec, pDir) <= 0 && normalVec.length() <= 2 * this.particleRadius) {
          normalVec.normalize();
          normalVec.scaleToRef(particle.direction.x * normalVec.x + particle.direction.y * normalVec.y + particle.direction.z * normalVec.z, pNormalVec, pNormalVec);
          normalVec.scaleToRef(q.direction.x * normalVec.x + q.direction.y * normalVec.y + q.direction.z * normalVec.z, qNormalVec);
          particle.direction.subtractToRef(pNormalVec, pTangVec);
          q.direction.subtractToRef(qNormalVec, qTangVec);
          pNormalVec.scaleInPlace(-1);
          pNormalVec.addToRef(pTangVec, particle.direction);
          particle.direction.normalize();
          particle.direction.scaleToRef(particle.speed, particle.velocity);
          qNormalVec.scaleInPlace(-1);
          qNormalVec.addToRef(qTangVec, q.direction);
          q.direction.normalize();
          q.direction.scaleToRef(q.speed, q.velocity);
          var pp = Math.floor(particle.idx / spheresPerShapeNb);
          for (var i = 0; i < spheresPerShapeNb; i++) {
            SPS.particles[spheresPerShapeNb * particleGroup + i].velocity = particle.velocity;
            SPS.particles[spheresPerShapeNb * qGroup + i].velocity = q.velocity;
          }
          p = spheresPerShapeNb * particleGroup + spheresPerShapeNb - 1;
        }
      }

      if (!setVelocity(this.leftNorm, this.leftPosition)) { //Left wall
        if (!setVelocity(this.rightNorm, this.rightPosition)) { //Right wall
          if (!setVelocity(this.frontNorm, this.frontPosition)) { //Front wall
            if (!setVelocity(this.backNorm, this.backPosition)) { //Back wall
              if (!setVelocity(this.groundNorm, this.groundPosition)) { //Ground
                setVelocity(this.roofNorm, this.roofPosition) //Roof
              }
            }
          }
        }
      };

      if (particle.idx % spheresPerShapeNb == spheresPerShapeNb - 1) {
        SPS.particles[spheresPerShapeNb * particleGroup].position.addInPlace(particle.velocity);
        SPS.particles[spheresPerShapeNb * particleGroup].centre = SPS.particles[spheresPerShapeNb * particleGroup].position;
        for (var i = 1; i < spheresPerShapeNb; i++) {
          SPS.particles[spheresPerShapeNb * particleGroup + i].centre = SPS.particles[spheresPerShapeNb * particleGroup].centre;
          SPS.particles[spheresPerShapeNb * particleGroup + i].position = SPS.particles[spheresPerShapeNb * particleGroup].position.add(SPS.particles[spheresPerShapeNb * particleGroup + i].offset);
          SPS.particles[spheresPerShapeNb * particleGroup + i].angle += SPS.particles[spheresPerShapeNb * particleGroup + i].angularSpeed;
          SPS.particles[spheresPerShapeNb * particleGroup + i].position = SPS.particles[spheresPerShapeNb * particleGroup + i].position.rotateAboutAxisThroughPoint(SPS.particles[spheresPerShapeNb * particleGroup + i].axis, SPS.particles[spheresPerShapeNb * particleGroup + i].centre, SPS.particles[spheresPerShapeNb * particleGroup + i].angle);
        }
      }
    }

    SPS.initParticles();

    this._scene.registerBeforeRender(function () {
      SPS.setParticles();
    });

  }

  animate(): void {
    var self = this;
    // run the render loop
    this._engine.runRenderLoop(() => {
      if (self._scene) 
        self._scene.render();
    });

    // the canvas/window resize event handler
    window.addEventListener('resize', () => {
      self._engine.resize();
    });
  }

}
